/*******************************************************************************
  Main Source File

  Company:
    Microchip Technology Inc.

  File Name:
    main.c

  Summary:
    This file contains the "main" function for a project.

  Description:
    This file contains the "main" function for a project.  The
    "main" function calls the "SYS_Initialize" function to initialize the state
    machines of all modules in the system
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include <stddef.h>                     // Defines NULL
#include <stdbool.h>                    // Defines true
#include <stdlib.h>                     // Defines EXIT_FAILURE
#include "definitions.h"                // SYS function prototypes
#include <math.h>

// *****************************************************************************
// *****************************************************************************
// Section: Main Entry Point
// *****************************************************************************
// *****************************************************************************
volatile float f_s = 10000;                                                     
volatile float two32 = 4294967296;                                              // 2^32 

volatile bool button_flag = false;                                              // Flag koji naznacava da je neki od tastera pritisnut - generisan je eksterni interrupt.
volatile bool tmr_flag = false;                                                 // Flag koji naznacava da je generisan tajmerski interrupt na svakih 0,1 ms.
volatile float f_out = 100;                                                     // Frekvencija izlazne sinusoide.

volatile uint32_t phase_accum;                                                  // Fazni akumulator.
volatile uint32_t N;                                                            // Inkrement faznog akumulatora.

// Look-up tabela sa vrednostima amplitude sinusoide 2048+2048*sin(2*pi*i/256)
const uint16_t sin_table[256] = {2048, 2098, 2148, 2199, 2249, 2299, 2349, 2398, 2448, 2497, 2546, 2594, 2643, 2690, 2738, 2785, 2832,
2878, 2924, 2969, 3013, 3057, 3101, 3144, 3186, 3227, 3268, 3308, 3347, 3386, 3423, 3460, 3496, 3531, 3565, 3599, 3631, 3663, 3693, 
3722, 3751, 3778, 3805, 3830, 3854, 3877, 3899, 3920, 3940, 3959, 3976, 3993, 4008, 4022, 4035, 4046, 4057, 4066, 4074, 4081, 4086, 
4090, 4094, 4095, 4095, 4095, 4094, 4090, 4086, 4081, 4074, 4066, 4057, 4046, 4035, 4022, 4008, 3993, 3976, 3959, 3940, 3920, 3899, 
3877, 3854, 3830, 3805, 3778, 3751, 3722, 3693, 3663, 3631, 3599, 3565, 3531, 3496, 3460, 3423, 3386, 3347, 3308, 3268, 3227, 3186, 
3144, 3101, 3057, 3013, 2969, 2924, 2878, 2832, 2785, 2738, 2690, 2643, 2594, 2546, 2497, 2448, 2398, 2349, 2299, 2249, 2199, 2148, 2098, 
2048, 1998, 1948, 1897, 1847, 1797, 1747, 1698, 1648, 1599, 1550, 1502, 1453, 1406, 1358, 1311, 1264, 1218, 1172, 1127, 1083, 1039, 995, 
952, 910, 869, 828, 788, 749, 710, 673, 636, 600, 565, 531, 497, 465, 433, 403, 374, 345, 318, 291, 266, 242, 219, 197, 176, 156, 137, 120, 
103, 88, 74, 61, 50, 39, 30, 22, 15, 10, 6, 2, 1, 0, 1, 2, 6, 10, 15, 22, 30, 39, 50, 61, 74, 88, 103, 120, 137, 156, 176, 197, 219, 242, 
266, 291, 318, 345, 374, 403, 433, 465, 497, 531, 565, 600, 636, 673, 710, 749, 788, 828, 869, 910, 952, 
995, 1039, 1083, 1127, 1172, 1218, 1264, 1311, 1358, 1406, 1453, 1502, 1550, 1599, 1648, 1698, 1747, 1797, 1847, 1897, 1948, 1998};

uint16_t dac_in;                                                                // Vrednost koja se vodi na D/A konvertor.
uint8_t data[2];

// Ukoliko je pritisnut taster RD0 generise se eksterni interrupt, proverava se da li je stanje na RD0 pinu logicka 0 i ukoliko jeste frekvencija izlazne sinusoide se smanjuje za 10.
void RD0_IntHandler(GPIO_PIN pin, uintptr_t context){
    if(GPIO_PinRead(pin)==0){
        f_out = f_out<=0 ? 0:f_out-10;                                          // Ukoliko je frekvencija smanjena na 0, ostaje na 0, sve dok se ne poveca.
        button_flag = true;
    }
}
// Ukoliko je pritisnut taster RD1 generise se eksterni interrupt, proverava se da li je stanje na RD1 pinu logicka 0 i ukoliko jeste frekvencija izlazne sinusoide se uvecava za 10.
void RD1_IntHandler(GPIO_PIN pin, uintptr_t context){
    if(GPIO_PinRead(pin)==0){
        f_out = f_out>=500 ? 500:f_out+10;                                      // Ukoliko je frekvencija povecana na 500, ostaje na 500, sve dok se ne smanji.
        button_flag = true;
    }
}

// Na svakih 0.1 ms koje izbroji TMR2 generise se interrupt - setuje se tmr_flag promenljiva:
void TMR2_IntHandler(uint32_t status, uintptr_t context){
    tmr_flag = true;
}


int main ( void )
{
    N =(uint32_t)(f_out*(two32/f_s)+0.5);                                       // Racunanje inkrementa faznog akumulatora.
    /* Initialize all modules */
    SYS_Initialize ( NULL );

    GPIO_PinInterruptCallbackRegister(GPIO_PIN_RD0, RD0_IntHandler, (uintptr_t)NULL);           // Registrovanje callback funkcije za eksterni interrupt na RD0.
    GPIO_PinInterruptEnable(GPIO_PIN_RD0);                                                      // Omogucavanje generisanja eksternog interrupt-a na RD0.
    
    GPIO_PinInterruptCallbackRegister(GPIO_PIN_RD1, RD1_IntHandler, (uintptr_t)NULL);           // Registrovanje callback funkcije za eksterni interrupt na RD1.
    GPIO_PinInterruptEnable(GPIO_PIN_RD1);                                                      // Omogucavanje generisanja eksternog interrupt-a na RD1.
    
    TMR2_CallbackRegister(TMR2_IntHandler, (uintptr_t)NULL);                    // Registrovanje callback funkcije za tajmerski interrupt na TMR2.
    TMR2_Start();                                                               // Startovanje TMR2.
    TMR2_PeriodSet(10000);                                                      // Postavljanje perioda na 10 000 - inkrement TMR2 se desava na svakih 1/f_clk = 1/100 MHz = 10 ns --> za 0,1 ms potrebo je da TMR2 broji 10 ns * 10 000 = 0,1 ms.
    

    while ( true )
    {
        if(tmr_flag){
            tmr_flag = false;                                                   // Ukoliko je proslo 0,1 ms generise se interrupt sa TMR2 - potrebno je resetovati tmr_flag.
            
            if(button_flag){                                                    // Ukoliko je pritisnut RD0 ili RD1 frekvencija f_out se menja, pa je potrebno preracunati i inkrement faznog akumulatora.
                button_flag = false;
                N =(uint32_t)(f_out*(two32/f_s)+0.5);
            }
            
            phase_accum+=N;                                                     // Vrednost faznog akumulatora inkrementira se za N.
            dac_in = sin_table[phase_accum>>24]|0x7000;                         // Look-up tabela se indeksira sa gornjim bajtom vrednosti faznog akumulatora - do te vrednosti se dolazi deljenjem faznog akumulatora sa 2^24 tj. siftovanjem za 24 bita u desno.
                                                                                // Dodatno, uz bitove podataka na D/A konvertor salju se i 4 konfiguraciona bita koja je potrebno postaviti na 0100 (odabir DAC_A, baferovanje ulaza, output gain = 1, output HiZ).
            data[0] = dac_in>>8;                                                // Na D/A konvertor salje se prvo high bajt podataka,
            data[1] = dac_in&0xFF;                                              // a zatim i low bajt podataka.

            GPIO_PinClear(GPIO_PIN_RF0);                                        // CS linija se obara na logicku 0.
            SPI5_Write(data, 2);                                                // Putem MOSI linije se na D/A konvertor salju podaci (4 konfiguraciona bita i podaci za konvertovanje).
            while(SPI5_IsTransmitterBusy());                                    // Dok traje slanje podataka kod se blokira.
            GPIO_PinSet(GPIO_PIN_RF0);                                          // CS linija se vraca u stanje logicke 1.

        }
    }

    /* Execution should not come here during normal operation */

    return ( EXIT_FAILURE );
}


/*******************************************************************************
 End of File
*/

